/*
 (MIT License)
 
 Copyright (c) 2004 NS Software
 
 Permission is hereby granted, free of charge, to any person obtaining a copy of this
 software and associated documentation files (the "Software"), to deal in the Software
 without restriction, including without limitation the rights to use, copy, modify,
 merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
 permit persons to whom the Software is furnished to do so, subject to the following
 conditions:
 
 The above copyright notice and this permission notice shall be included in all copies
 or substantial portions of the Software.
 
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
 INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
 OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
 */
/*
 *  Fragment.cpp
 *  Racer
 *
 *  Created by Nathan Sturtevant on Sat Apr 10 2004.
 *  Copyright (c) 2004 NS Software.
 *
 */

#include "Fragment.h"
#include "point3d.h"
#include "vector3d.h"

Fragment::Fragment(GLfloat posx, GLfloat posy, GLfloat posz,
				 GLfloat velx, GLfloat vely, GLfloat velz,
				 GLfloat rotx, GLfloat roty)
:Object(kDisplayObject, 0)
{
	pos[0] = posx;
	pos[1] = posy;
	pos[2] = posz;
	velocity[0] = velx;
	velocity[1] = vely;
	velocity[2] = velz;
	rotation[0] = rotx;
	rotation[1] = roty;
	elapsedTime = 0;
}

void Fragment::draw()
{
	glColor3f(1.0-elapsedTime/3, 1.0-elapsedTime/3, 1.0-elapsedTime/3);
	glPushMatrix();
	glTranslatef(pos[0], pos[1], pos[2]);
	glRotatef(rotation[0], 1.0, 0.0, 0.0);
	glRotatef(rotation[1], 0.0, 1.0, 0.0);

	glBegin(GL_LINES);
	glVertex3f(0.02, 0.02, 0.02);
	glVertex3f(-0.02, -0.02, -0.02);
	glEnd();
	 
	glPopMatrix();
}

uint32_t Fragment::stepTime(double amount)
{
	elapsedTime += amount;
	rotation[0]+=1;
	rotation[1]-=1;
	for (int x = 0; x < 3; x++)
		pos[x]+=amount*velocity[x];
	if (elapsedTime >= 4)
		return kObjectDestroyed;
	return kObjectRemains;
}

point3d Fragment::getLocation()
{
	point3d p;
	p.x = pos[0];
	p.y = pos[1];
	p.y = pos[2];
	return p;
}
