/*
 (MIT License)
 
 Copyright (c) 2004 NS Software
 
 Permission is hereby granted, free of charge, to any person obtaining a copy of this
 software and associated documentation files (the "Software"), to deal in the Software
 without restriction, including without limitation the rights to use, copy, modify,
 merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
 permit persons to whom the Software is furnished to do so, subject to the following
 conditions:
 
 The above copyright notice and this permission notice shall be included in all copies
 or substantial portions of the Software.
 
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
 INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
 OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
 */
/*
 *  Music.cpp
 *  Racer
 *
 *  Created by Nathan Sturtevant on Sat Apr 10 2004.
 *  Copyright (c) 2004 NS Software.
 *
 */

#include "Music.h"

Music::Music(int *_score, int _len, int _options, uint32_t _parameter, int _instrument, bool _se)
//Music::Music(int offset)
:Object(kDisplayObject, 0)
{
	ComponentResult thisError;
	elapsedTime = 0;

	done = false;
	score = 0;
	na = 0;
	nc = 0;
	currentNote = 0;
		offset = 0;
    soundEffect = _se;
	// Open up the note allocator.
	na = OpenDefaultComponent('nota'/*kNoteAllocatorType*/, 0);
	if (!na)
	{
		printf("Failed to open note component\n");
		return;
	}
	
	// Fill out a NoteRequest using NAStuffToneDescription to help, and
	// allocate a NoteChannel.
	nr.info.flags = 0;
	nr.info.midiChannelAssignment = 0;
	nr.info.polyphony = 2; // simultaneous tones
	nr.info.typicalPolyphony = 0x00010000; // usually just one note
	thisError = NAStuffToneDescription(na, _instrument, &nr.tone); // 1 is piano
	thisError = NANewNoteChannel(na, &nr, &nc);

	speed = 1;
	length = _len;
	score = new int[length];
	for (int x = 0; x < length; x++)
		score[x] = _score[x];
	maxTimesPlayed = 0;
	timesRepeated = 0;
	timesPlayed = 0;
	speedFactor = 1.0;
	repeatDir = 1;
	if (_options == kPlayOnce)
		fReturnValue = kObjectDestroyed;
	else
		fReturnValue = kObjectDisabled;
	if (_options == kRepeat)
	{
		maxTimesPlayed = _parameter;
	}
	if ((_options == kRepeatRisingSpeedup) || (_options == kRepeatSpeedup))
	{
		speedFactor = .99;
		repeatDir = 1;
	}
	if ((_options == kRepeatRisingSlowdown) || (_options == kRepeatSlowdown))
	{
		speedFactor = 1.01;
		repeatDir = -1;
	}
	if ((_options == kRepeatRisingSlowdown) || (_options == kRepeatSlowdown) ||
			(_options == kRepeatRisingSpeedup) || (_options == kRepeatSpeedup))
	{
		maxTimesPlayed = (_parameter&0xFFFF)*((_parameter>>16)&0xFFFF);
		timesRepeated = (_parameter&0xFFFF);
	}
}

Music::~Music()
{
	NAResetNoteChannel(na, nc);
	if (nc)
		NADisposeNoteChannel(na, nc);
	if (na)
		CloseComponent(na);
	delete [] score;
}

// score is in pairs of note & duration (60ths of a second)
uint32_t Music::stepTime(double amount)
{
	if ((done == true) || (!na) || (!nc))
	{
		//NAResetNoteChannel(na, nc);
		return kObjectDestroyed;
	}
	
	if (amount == 0)
	{
		NAResetNoteChannel(na, nc);
		return kObjectRemains;
	}
	
	elapsedTime += amount;
//	printf("On note %d (%d/%d), current time %1.2f\n", currentNote, score[currentNote],
//				 score[currentNote+1], elapsedTime);
	if (elapsedTime > speed*score[currentNote+1]/60.0)
	{
		//printf("Ending note %d\n", score[currentNote]+(timesPlayed/timesRepeated)*repeatDir);
		// end this note, start the next
		NAPlayNote(na, nc, score[currentNote]+(timesPlayed/timesRepeated)*repeatDir+offset, 0);
		currentNote+=2;
		if (currentNote >= length)
		{
			timesPlayed++;
			currentNote = 0;
			speed *= speedFactor;
			if (timesPlayed > maxTimesPlayed) 
			{// if we are repeating, this keeps us around, in case someone has a pointer to us
				return fReturnValue;
			}
		}
		elapsedTime = 0;
	}
	if (elapsedTime == 0)
	{
		if (score[currentNote] != 0)
		{
      if (((!soundEffect) && (gMusic)) || ((soundEffect) && (gSound)))
        NAPlayNote(na, nc, score[currentNote]+(timesPlayed/timesRepeated)*repeatDir+offset, 80);
			//printf("Starting note %d\n", score[currentNote]+(timesPlayed/timesRepeated)*repeatDir);
		}
	}
	return kObjectRemains;
}

void Music::setPitchOffset(int val)
{
//	NAResetNoteChannel(na, nc);
	offset = val;
}

void Music::destroy()
{
	done = true;
}
