/*
 (MIT License)
 
 Copyright (c) 2004 NS Software
 
 Permission is hereby granted, free of charge, to any person obtaining a copy of this
 software and associated documentation files (the "Software"), to deal in the Software
 without restriction, including without limitation the rights to use, copy, modify,
 merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
 permit persons to whom the Software is furnished to do so, subject to the following
 conditions:
 
 The above copyright notice and this permission notice shall be included in all copies
 or substantial portions of the Software.
 
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
 INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
 OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
 */
/*
 *  waygate.cpp
 *  Racer
 *
 *  Created by Nathan Sturtevant on Thu Apr 08 2004.
 *  Copyright (c) 2004 NS Software.
 *
 */

#include "WayGate.h"

WayGate::WayGate(GLfloat locx, GLfloat locy, GLfloat locz, GLfloat rotx, GLfloat roty, GLfloat rotz, double _radius, int resolution)
	:Object(kWaygateObject, kBulletObject|kHomingMineObject|kShipObject)
{
		res = resolution;
		elapsedTime = 0;
		radius = _radius;
		pos[0] = locx;
		pos[1] = locy;
		pos[2] = locz;
		rot[0] = rotx;
		rot[1] = roty;
		rot[2] = rotz;
		//passedMusic = 0;
		
		n1.x = pos[0]+sin(roty*PI/180)*cos(rotx*PI/180)*cos(rotz*PI/180)+sin(rotx*PI/180)*sin(rotz*PI/180);
		n1.y = pos[1]+sin(roty*PI/180)*sin(rotz*PI/180)*cos(rotx*PI/180)-sin(rotx*PI/180)*cos(rotz*PI/180);
		n1.z = pos[2]+cos(rotx*PI/180)*cos(roty*PI/180);
		n2.x = pos[0]+-sin(roty*PI/180)*cos(rotx*PI/180)*cos(rotz*PI/180)-sin(rotx*PI/180)*sin(rotz*PI/180);
		n2.y = pos[1]+-sin(roty*PI/180)*sin(rotz*PI/180)*cos(rotx*PI/180)+sin(rotx*PI/180)*cos(rotz*PI/180);
		n2.z = pos[2]+-cos(rotx*PI/180)*cos(roty*PI/180);

		locked = false;
}

void WayGate::lock()
{
	if (!locked)
	{
		if (elapsedTime > 1)
		{
			int notes[4] = {0, 1, 60, 30};
			pushCreatedObject(new Music(notes, 4, kPlayOnce, 0, 128*1+122));
		}
		locked = true;
	}
}

void WayGate::unlock()
{
	locked = false;
}

void WayGate::draw()
{
	glPushMatrix();
  float dRes = gDetail?(res*4):res;
//	glColor3f(0.0, 0.0, 1.0);
//	glBegin(GL_LINES);
//	glVertex3f(n1.x, n1.y, n1.z);
//	glVertex3f(n2.x, n2.y, n2.z);
//	glEnd();

	glTranslatef(pos[0], pos[1], pos[2]);
	glRotatef(rot[0], 1.0, 0.0, 0.0);
	glRotatef(rot[1], 0.0, 1.0, 0.0);
	glRotatef(rot[2], 0.0, 0.0, 1.0);

	glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);

	glColor3f(0.0, 1.0, 0.0);
	glBegin(GL_QUAD_STRIP);
	for (double x = 0; x <= TWOPI+.001; x += TWOPI/dRes)
	{
		double xoff = cos(x);
		double yoff = sin(x);
		glVertex3f(xoff*radius, yoff*radius, -.1);
		glVertex3f(xoff*(radius+radius/10), yoff*(radius+radius/10), 0);
	}
	glEnd();
	glBegin(GL_QUAD_STRIP);
	for (double x = 0; x <= TWOPI+.001; x += TWOPI/dRes)
	{
		double xoff = cos(x);
		double yoff = sin(x);
		glVertex3f(xoff*(radius+radius/10), yoff*(radius+radius/10), 0);
		glVertex3f(xoff*radius, yoff*radius, .1);
	}
	glEnd();
	glBegin(GL_QUAD_STRIP);
	for (double x = 0; x <= TWOPI+.001; x += TWOPI/dRes)
	{
		double xoff = cos(x);
		double yoff = sin(x);
		glVertex3f(xoff*radius, yoff*radius, -.1);
		glVertex3f(xoff*radius, yoff*radius, .1);
	}
	glEnd();
	
	
	glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
	glColor3f(0.0, 0.0, 0.0);
	glBegin(GL_QUAD_STRIP);
	for (double x = 0; x <= TWOPI+.001; x += TWOPI/dRes)
	{
		double xoff = cos(x);
		double yoff = sin(x);
		glVertex3f(xoff*radius, yoff*radius, -.1);
		glVertex3f(xoff*(radius+radius/10), yoff*(radius+radius/10), 0);
	}
	glEnd();
	glBegin(GL_QUAD_STRIP);
	for (double x = 0; x <= TWOPI+.001; x += TWOPI/dRes)
	{
		double xoff = cos(x);
		double yoff = sin(x);
		glVertex3f(xoff*(radius+radius/10), yoff*(radius+radius/10), 0);
		glVertex3f(xoff*radius, yoff*radius, .1);
	}
	glEnd();
	glBegin(GL_QUAD_STRIP);
	for (double x = 0; x <= TWOPI+.001; x += TWOPI/dRes)
	{
		double xoff = cos(x);
		double yoff = sin(x);
		glVertex3f(xoff*radius, yoff*radius, -.1);
		glVertex3f(xoff*radius, yoff*radius, .1);
	}
	glEnd();
	
	if (locked)
	{
		glColor4f(1.0, 0.0, 0.0, 1.0);
		glBegin(GL_LINES);
		for (double x = TWOPI/dRes; x < TWOPI; x += TWOPI/dRes)
		{
			//glVertex3f(0, 0, 0);
			double xoff = sin(x+elapsedTime/4);
			double yoff = -cos(x+elapsedTime/4);
      //if (((int)elapsedTime)%2)
      {
        glVertex3f(xoff*radius, yoff*radius, 0);
        glVertex3f(-xoff*radius, yoff*radius, 0);
      }
//      else {
        glVertex3f(yoff*radius, xoff*radius, 0);
        glVertex3f(yoff*radius, -xoff*radius, 0);
//      }
		}
		glEnd();
	}

	if (activeGoal)
	{
    glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
		glColor3f(0.0, 0.5, 0.0);
		glBegin(GL_TRIANGLE_FAN);
		glVertex3f(0, 0, 0);
		for (double x = 0; x <= TWOPI; x += TWOPI/4)
		{
			double xoff = cos(x+elapsedTime/2);
			double yoff = sin(x+elapsedTime/2);
			glVertex3f(xoff*radius, yoff*radius, 0);
		}
		glEnd();
		glBegin(GL_TRIANGLE_FAN);
		glVertex3f(0, 0, 0);
		for (double x = 0; x <= TWOPI; x += TWOPI/4)
		{
			double xoff = cos(x-elapsedTime/2);
			double yoff = sin(x-elapsedTime/2);
			glVertex3f(xoff*radius, yoff*radius, 0);
		}
		glEnd();
    glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
	}
	
	glPopMatrix();
}

uint32_t WayGate::stepTime(double amount)
{
	elapsedTime += amount;
	//if (passedMusic)
	//return kObjectRemains|kNewObjectsCreated;
	return kObjectRemains;
}


point3d WayGate::getLocation()
{
	point3d p;
	p.x = pos[0];
	p.y = pos[1];
	p.z = pos[2];
	return p;
}

uint32_t WayGate::testCollision(Object *o)
{
	point3d p1 = o->getLocation();
	// ignore if we are more than radius away from center...
	double cDist = getLocation().getDistance(p1);
	
	if (cDist > getRadius()+o->getRadius())
		return kNothingDestroyed;
	
	double dist1 = n1.getDistance(p1);
	double dist2 = n2.getDistance(p1);
	// if the distance between the normals is within the size of your object, we've crossed
	// the plane
	if (fabs(dist1 - dist2) < o->getRadius())
	{
		// if the distance is such that it puts us in the edge of the gate,
		// we get destroyed
		if (cDist+o->getRadius() > radius)
		{
			//printf("Object hit waygate\n");
			return kThatDestroyed;
		}
		if (o->getType() == kShipObject)
		{
			if (locked)
			{
				return kThatDestroyed;
			}
			if (activeGoal)
			{
				int notes[18] = {0, 0, 66, 12, 66, 0, 70, 12, 70, 0, 73, 12, 73, 0, 78, 12, 78, 0};
				pushCreatedObject(new Music(notes, 18, kPlayOnce, 0, 91));
				activeGoal = false;
				if (nextGoal)
					nextGoal->activate();
			}
		}
	}
		
		//	if (p1.length() < getRadius()+o->getRadius())
//	{
//		return kThisDestroyed;
//	}
	
	// figure this out later...
	return kNothingDestroyed;
	//	kThisDestroyed,
	//	kThatDestroyed,
	//	kBothDestroyed
}

//Object *WayGate::getNewObject()
//{
//	if (passedMusic)
//	{
//		Music *tmp = passedMusic;
//		passedMusic = 0;
//		return tmp;
//	}
//	return 0;
//}

void WayGate::destroy()
{
	// waygates are indestructible!
}

void WayGate::setColor()
{
	if (locked)
	{
		if (activeGoal)
			glColor3f(.5, 1, 0);
		else
			glColor3f(.5, .6, 0);
	}
	else {
		if (activeGoal)
			glColor3f(0, 1, 0);
		else
			glColor3f(0, .6, 0);
	}
}
