/*
 (MIT License)
 
 Copyright (c) 2004 NS Software
 
 Permission is hereby granted, free of charge, to any person obtaining a copy of this
 software and associated documentation files (the "Software"), to deal in the Software
 without restriction, including without limitation the rights to use, copy, modify,
 merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
 permit persons to whom the Software is furnished to do so, subject to the following
 conditions:
 
 The above copyright notice and this permission notice shall be included in all copies
 or substantial portions of the Software.
 
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
 INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
 OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

 */

#define kUseMultiSample 0

#include <OpenGL/glu.h>
#include <OpenGL/glext.h>

#include "HIDSupport.h"
#include "glCheck.h"
#include "trackball.h"
#include "main.h"
#include "World.h"
#include "Stars.h"
#include "Bullet.h"
#include "Mine.h"
#include "WayGate.h"
#include "Music.h"
#include "YourShip.h"
#include "HomingMine.h"
#include "TextBox.h"
#include "GateKey.h"
#include "Turret.h"

void RegisterHelpBook();
OSStatus disposeGL(pRecContext pContextInfo);
void loadLevel(pRecContext pContextInfo);
void setupMainMenu(pRecContext pContextInfo);
void updateScoreTime(pRecContext pContextInfo);
void resetNewGame(pRecContext pContextInfo);
void handleMainMenuClick(pRecContext pContextInfo, HIPoint location, Rect r);
void handleWaitingClick(pRecContext pContextInfo);
void handleEnterHighClick(pRecContext pContextInfo, HIPoint location, Rect r);
void handleBetweenLevelClick(pRecContext pContextInfo, Rect r);
void handlePausedClick();
void handleGameClick(pRecContext pContextInfo, EventMouseButton button, UInt32 modifiers);
void handleGameMouseMovement(pRecContext pContextInfo, HIPoint location, Rect r);
void loadLevelDescription(pRecContext pContextInfo);
void updateRadarZoom(pRecContext pContextInfo);
bool isPressed(unsigned short k);

void LoadPrefs(pRecContext pContextInfo);
void SavePrefs(pRecContext pContextInfo);
void setPrefValue(CFStringRef key, int value);
int getPrefValue(CFStringRef key);


// ==================================

recVec gOrigin = {0.0, 0.0, 0.0};
static bool first = true;

// single set of interaction flags and states
GLint gDollyPanStartPoint[2] = {0, 0};
GLfloat gTrackBallRotation [4] = {0.0f, 0.0f, 0.0f, 0.0f};
GLboolean gDolly = GL_FALSE;
GLboolean gPan = GL_FALSE;
GLboolean gTrackball = GL_FALSE;
pRecContext gTrackingContextInfo = NULL;
GLfloat gRadius = 64.0;

int gMusic=1, gSound=1, gDetail=0, gInvert=0;

enum
{
	kMainMenu = 500,
	kCloseMenuItem = 2,
	kInfoMenuItem = 4,
	kAnimateMenuItem = 5,
	kInfoState = 1,
	kAnimateState = 1
};

EventHandlerUPP gEvtHandler;			// main event handler
EventHandlerUPP gWinEvtHandler;			// window event handler

IBNibRef nibRef = NULL;

AbsoluteTime gStartTime;

char gErrorMessage[256] = ""; // buffer for error message output
float gErrorTime = 0.0;

#pragma mark ---- OpenGL Capabilities ----

// GL configuration info globals
// see glcheck.h for more info
GLCaps * gDisplayCaps = NULL; // array of GLCaps
CGDisplayCount gNumDisplays = 0;
// related DM change notification:
DMExtendedNotificationUPP gConfigEDMUPP = NULL;

static void getCurrentCaps(void)
{
	// Check for existing opengl caps here
	// This can be called again with same display caps array when display configurations are changed and
	//   your info needs to be updated.  Note, if you are doing dynmaic allocation, the number of displays
	//   may change and thus you should always reallocate your display caps array.
	if (gDisplayCaps && HaveOpenGLCapsChanged(gDisplayCaps, gNumDisplays)) { // see if caps have changed
		free(gDisplayCaps);
		gDisplayCaps = NULL;
	}
	if (!gDisplayCaps) { // if we do not have caps
		CheckOpenGLCaps(0, NULL, &gNumDisplays); // will just update number of displays
		gDisplayCaps =(GLCaps*) malloc(sizeof(GLCaps) * gNumDisplays);
		CheckOpenGLCaps(gNumDisplays, gDisplayCaps, &gNumDisplays);
	}
}

#pragma mark ---- Utilities ----

// return float elpased time in seconds since app start
static float getElapsedTime(void)
{	
	float deltaTime =(float) AbsoluteDeltaToDuration(UpTime(), gStartTime);
	if (0 > deltaTime)	// if negative microseconds
		deltaTime /= -1000000.0;
	else				// else milliseconds
		deltaTime /= 1000.0;
	return deltaTime;
}

#pragma mark ---- Error Reporting ----

// C string to Pascal string
static void cstr2pstr(StringPtr outString, const char *inString)
{	
	unsigned short x = 0;
	do {
		outString [x + 1] =(unsigned char) inString [x];
		x++;
	} while ((inString [x] != 0)  &&(x < 256));
	outString [0] = x;									
}

// ---------------------------------

// error reporting as both window message and debugger string
void reportError(char * strError)
{
	Str255 strErr = "\p";
	
	gErrorTime = getElapsedTime();
	sprintf(gErrorMessage, "Error: %s(at time: %0.1f secs)", strError, gErrorTime);
	
	// out as debug string
	cstr2pstr(strErr, gErrorMessage);
	DebugStr(strErr);
}

// ---------------------------------

// if error dump agl errors to debugger string, return error
OSStatus aglReportError(void)
{
	GLenum err = aglGetError();
	if (AGL_NO_ERROR != err)
		reportError((char *) aglErrorString(err));
	// ensure we are returning an OSStatus noErr if no error condition
	if (err == AGL_NO_ERROR)
		return noErr;
	else
		return(OSStatus) err;
}

// ---------------------------------

// if error dump gl errors to debugger string, return error
OSStatus glReportError(void)
{
	GLenum err = glGetError();
	if (GL_NO_ERROR != err)
		reportError((char *) gluErrorString(err));
	// ensure we are returning an OSStatus noErr if no error condition
	if (err == GL_NO_ERROR)
		return noErr;
	else
		return(OSStatus) err;
}

#pragma mark ---- OpenGL Utilities ----

// C string drawing function
void drawCStringGL(char * cstrOut, GLuint fontList)
{
	GLint i = 0;
	while (cstrOut [i])
		glCallList(fontList + cstrOut[i++]);
}

// ---------------------------------

// AGL bitmpp font setup
GLuint buildFontGL(AGLContext ctx, GLint fontID, Style face, GLint size)
{
	GLuint listBase = glGenLists(256);
	if (aglUseFont(ctx, fontID , face, size, 0, 256,(long) listBase)) {
		glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
		glReportError();
		return listBase;
	} else {
		reportError("aglUseFont failed\n" );
		glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
		glDeleteLists(listBase, 256);
		return 0;
	}
}

// ---------------------------------

// delete font list passed in
void deleteFontGL(GLuint fontList)
{
	if (fontList)
		glDeleteLists(fontList, 256);
}

// ---------------------------------

// update the projection matrix based on camera and view info
// should be called when viewport size, eye z position, or camera aperture changes
// also call if far or near changes which is determined by shape size in this case
void updateProjection(pRecContext pContextInfo)
{
	GLdouble ratio, radians, wd2;
	GLdouble left, right, top, bottom, near, far;
	
	aglSetCurrentContext(pContextInfo->aglContext);
	
	// set projection
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	far = 100;
	near = 0.2;
	radians = 0.0174532925 * pContextInfo->camera.aperture / 2; // half aperture degrees to radians 
	wd2 = near * tan(radians);
	ratio = 1;
	//ratio = pContextInfo->camera.viewWidth /(float) pContextInfo->camera.viewHeight;
	if (ratio >= 1.0) {
		left  = -ratio * wd2;
		right = ratio * wd2;
		top = wd2;
		bottom = -wd2;	
	}
	else {
		left  = -wd2;
		right = wd2;
		top = wd2 / ratio;
		bottom = -wd2 / ratio;	
	}

	glFrustum(left, right, bottom, top, near, far);
}

// ---------------------------------

// updates the contexts model view matrix for object and camera moves
// we will call this every draw loop for simplicity
void updateModelView(pRecContext pContextInfo, bool overview)
{
	aglSetCurrentContext(pContextInfo->aglContext);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	if (overview)
	{
		if (pContextInfo->controlShip)
			pContextInfo->controlShip->setupOverviewpoint();
	}
	else {
		if (pContextInfo->controlShip)
			pContextInfo->controlShip->setupViewpoint();
	}
}

// ---------------------------------

// handles resizing of GL need context update and if the window dimensions change, a
// a window dimension update, reseting of viewport and an update of the projection matrix

OSStatus resizeGL(pRecContext pContextInfo, CGRect viewRect)
{
	OSStatus err = noErr;
	
	if (!pContextInfo || !(pContextInfo->aglContext))
		return paramErr;
	
	if (!aglSetCurrentContext(pContextInfo->aglContext)) err = aglReportError();
	if (!aglUpdateContext(pContextInfo->aglContext)) err = aglReportError();
	
	// ensure camera knows size changed
	//	if ((pContextInfo->camera.viewHeight != viewRect.size.height) ||
	//	   (pContextInfo->camera.viewWidth != viewRect.size.width)) {
	pContextInfo->camera.viewOriginX = viewRect.origin.x;
	pContextInfo->camera.viewOriginY = viewRect.origin.y;
	pContextInfo->camera.viewWidth =(GLint)viewRect.size.width;
	pContextInfo->camera.viewHeight =(GLint)viewRect.size.height;

	updateProjection(pContextInfo);  // update projection matrix
																		//	}
	return err;
}

// ---------------------------------

// sets the camera data to initial conditions
static void resetCamera(recCamera * pCamera)
{
	pCamera->aperture = 20;
	//pCamera->rotPoint = gOrigin;
	// will be set in resize once the target view size and position is known
	pCamera->viewOriginY = 0;
	pCamera->viewOriginX = 0;
	pCamera->viewHeight = 0;
	pCamera->viewWidth = 0;
}


#pragma mark ---- OpenGL Minimize Handler(thanks to Dan Herman) ----

void InvertGLImage( char *imageData, long imageSize, long rowBytes )
{
	long i, j;
	// This is not an optimized routine!
	char *tBuffer =(char*) malloc((size_t)rowBytes );
	if (NULL == tBuffer)
	{
		printf("Error mallocing image data\n");
		return;
	}
	
	// Copy rows into tmp buffer one at a time, reversing their order
	for(i = 0, j = imageSize - rowBytes; i < imageSize; i += rowBytes, j -= rowBytes) {
		memcpy( tBuffer, &imageData[i], rowBytes );
		memcpy( &imageData[i], &imageData[j], rowBytes );
		memcpy( &imageData[j], tBuffer, rowBytes);
		//memcpy( &tBuffer[j], &imageData[i],(size_t) rowBytes );
	}
	
	// Copy tmp buffer back into original buffer
	//memcpy( imageData, tBuffer,(size_t)imageSize );
	free(tBuffer);
}

// ---------------------------------

void CompositeGLBufferIntoWindow(AGLContext ctx, Rect *bufferRect, WindowRef win)
{
	GWorldPtr pGWorld;
	QDErr err;
	// blit OpenGL content into window backing store
	// allocate buffer to hold pane image
	long width  =(bufferRect->right - bufferRect->left);
	long height  =(bufferRect->bottom - bufferRect->top);
	Rect src_rect = {0, 0, height, width};
	long rowBytes = width * 4;
	long imageSize = rowBytes * height;
	char *image =(char *) NewPtr(imageSize);
	if (!image) {
		printf("Out of memory in CompositeGLBufferIntoWindow()!");
		return;		// no harm in continuing
	}
	
	// pull GL content down to our image buffer
	aglSetCurrentContext( ctx );
	glReadPixels(0, 0, width, height, GL_BGRA, GL_UNSIGNED_INT_8_8_8_8_REV, image);
	
	// GL buffers are upside-down relative to QD buffers, so we need to flip it
	InvertGLImage( image, imageSize, rowBytes );
	
	// create a GWorld containing our image
	err = NewGWorldFromPtr(&pGWorld, k32ARGBPixelFormat, &src_rect, 0, 0, 0, image, rowBytes);
	if (err != noErr) {
		printf("WARNING: error in NewGWorldFromPtr, called from CompositeGLBufferIntoWindow()");
		free( image );
		return;
	}
	
	SetPort( GetWindowPort(win));
	CopyBits( GetPortBitMapForCopyBits(pGWorld), GetPortBitMapForCopyBits(GetWindowPort(win)), &src_rect, bufferRect, srcCopy, 0 );
	DisposeGWorld( pGWorld );
	DisposePtr( image );
}

#pragma mark ---- Display Manager Event Handling ----

// update our GL configuration info based on display change notification
void handleConfigDMEvent(void *userData, short theMessage, void *notifyData)
{
	if (kDMNotifyEvent == theMessage) { // post change notifications only
		getCurrentCaps();
	}
}

// ---------------------------------

// handle display config changes meaing we need to update the GL context via the resize function and check for windwo dimension changes
// also note we redraw the content here as it could be lost in a display config change
void handleWindowDMEvent(void *userData, short theMessage, void *notifyData)
{
	if (kDMNotifyEvent == theMessage) { // post change notifications only
		pRecContext pContextInfo = NULL;
		WindowRef window =(WindowRef) userData;
		if (window)
			pContextInfo =(pRecContext) GetWRefCon(window);
		if (pContextInfo) { // have a valid OpenGl window
			Rect rectPort;
			CGRect viewRect = {{0.0f, 0.0f}, {0.0f, 0.0f}};
			sprintf(pContextInfo->message, "Event: Display Change at %0.1f secs", getElapsedTime());
			pContextInfo->msgTime = getElapsedTime();
			GetWindowPortBounds(window, &rectPort);
			viewRect.size.width =(float)(rectPort.right - rectPort.left);
			viewRect.size.height =(float)(rectPort.bottom - rectPort.top);
			resizeGL(GetCurrentContextInfo(window), viewRect); // update context and handle possible resizes
			InvalWindowRect(window,  &rectPort); // force redrow
		}
	}
}

#pragma mark ---- OpenGL setup and teardown ----

void SetLighting(unsigned int mode)
{
	GLfloat mat_specular[] = {1.0, 1.0, 1.0, 1.0};
	GLfloat mat_shininess[] = {90.0};
	
	GLfloat position[4] = {7.0,-7.0,12.0,0.0};
	GLfloat ambient[4]  = {0.2,0.2,0.2,1.0};
	GLfloat diffuse[4]  = {1.0,1.0,1.0,1.0};
	GLfloat specular[4] = {1.0,1.0,1.0,1.0};
	
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, mat_specular);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, mat_shininess);
	
	glEnable(GL_COLOR_MATERIAL);
	glColorMaterial(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE);
	
	switch(mode) {
		case 0:
			break;
		case 1:
			glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,GL_FALSE);
			glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER,GL_FALSE);
			break;
		case 2:
			glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,GL_FALSE);
			glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER,GL_TRUE);
			break;
		case 3:
			glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,GL_TRUE);
			glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER,GL_FALSE);
			break;
		case 4:
			glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,GL_TRUE);
			glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER,GL_TRUE);
			break;
	}
	
	glLightfv(GL_LIGHT0,GL_POSITION,position);
	glLightfv(GL_LIGHT0,GL_AMBIENT,ambient);
	glLightfv(GL_LIGHT0,GL_DIFFUSE,diffuse);
	glLightfv(GL_LIGHT0,GL_SPECULAR,specular);
	glEnable(GL_LIGHT0);
}

// ---------------------------------

OSStatus buildGL(WindowRef window)
{
	OSStatus err = noErr;
	Rect rectPort;
#ifndef kUseMultiSample
	GLint attrib[] = { AGL_RGBA, AGL_DOUBLEBUFFER, AGL_DEPTH_SIZE, 16, AGL_NONE };
#else
	GLint attrib[] = { AGL_RGBA, AGL_DOUBLEBUFFER, AGL_DEPTH_SIZE, 16, AGL_SAMPLE_BUFFERS_ARB, 1, AGL_SAMPLES_ARB, kSamples, AGL_NO_RECOVERY, AGL_NONE };
#endif
	pRecContext pContextInfo = GetCurrentContextInfo(window);
	ProcessSerialNumber psn = { 0, kCurrentProcess };
	
	if (NULL == pContextInfo)
		return paramErr;
	
	// build context
	pContextInfo->aglContext = NULL;
	pContextInfo->aglPixFmt = aglChoosePixelFormat(NULL, 0, attrib);
	aglReportError();
	if (pContextInfo->aglPixFmt) {
		pContextInfo->aglContext = aglCreateContext(pContextInfo->aglPixFmt, NULL);
		aglReportError();
	}
#if kUseMultiSample
  // if multi-sample pixel format or context fails allocate a non-multi-sample one
  if ((noErr != err) || (NULL == pContextInfo->aglContext)) { // try non-multisample
    attrib[5] = attrib[6] = attrib[7] = attrib[8] = 0;
    if (pContextInfo->aglPixFmt)
      aglDestroyPixelFormat (pContextInfo->aglPixFmt);
    pContextInfo->aglPixFmt = aglChoosePixelFormat(NULL, 0, attrib);
    err = aglReportError ();
    if (pContextInfo->aglPixFmt) {
      pContextInfo->aglContext = aglCreateContext (pContextInfo->aglPixFmt, NULL);
      err = aglReportError ();
    }
  }
#endif
	if (pContextInfo->aglContext) {
		short fNum;
		GLint swap = 1;
		CGRect viewRect = {{0.0f, 0.0f}, {0.0f, 0.0f}};
		
		GrafPtr portSave = NULL;
		GetPort(&portSave);
		SetPort((GrafPtr) GetWindowPort(window));
		
		if (!aglSetDrawable(pContextInfo->aglContext, GetWindowPort(window)))
			err = aglReportError();
		if (!aglSetCurrentContext(pContextInfo->aglContext))
			err = aglReportError();
		
		// VBL SYNC
		if (!aglSetInteger(pContextInfo->aglContext, AGL_SWAP_INTERVAL, &swap))
			aglReportError();
		
		switch(pContextInfo->modeFSAA) {
			case kFSAAOff:
				glDisable(GL_MULTISAMPLE_ARB);
				break;
			case kFSAAFast:
				glEnable(GL_MULTISAMPLE_ARB);
				glHint(GL_MULTISAMPLE_FILTER_HINT_NV, GL_FASTEST);
				break;
			case kFSAANice:
				glEnable(GL_MULTISAMPLE_ARB);
				glHint(GL_MULTISAMPLE_FILTER_HINT_NV, GL_NICEST);
				break;
		}
		
		// ensure we know when display configs are changed
		pContextInfo->windowEDMUPP = NewDMExtendedNotificationUPP(handleWindowDMEvent); // for display change notification
		DMRegisterExtendedNotifyProc(pContextInfo->windowEDMUPP,(void *) window, NULL, &psn);
		
		// init GL stuff here
		glEnable(GL_DEPTH_TEST);

		GLfloat fogColor[4] = {0, 0, 0, 0.5};
		glEnable(GL_FOG);
		glFogi(GL_FOG_MODE, GL_LINEAR);
		glFogf(GL_FOG_DENSITY, .2);
		glFogfv(GL_FOG_COLOR, fogColor);
		glHint(GL_FOG_HINT, GL_DONT_CARE);
		glFogf(GL_FOG_START, 50.0);
		glFogf(GL_FOG_END, 100.0);

    //glHint(GL_POINT_SMOOTH_HINT, GL_FASTEST);
    glDisable(GL_POINT_SMOOTH);
    glEnable(GL_LINE_SMOOTH);
		//glShadeModel(GL_SMOOTH);
		glShadeModel(GL_FLAT);
		glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
		glFrontFace(GL_CCW);
		glPolygonOffset(1.0, 1.0);
		
		glClearColor(0.0,0.0,0.0,0.0);
		//pContextInfo->shapeSize = 3.0; // max radius of of objects
		
		GetFNum("\pGeneva", &fNum); // build font
		pContextInfo->boldFontList = buildFontGL(pContextInfo->aglContext, fNum, bold, 9);
		pContextInfo->regFontList = buildFontGL(pContextInfo->aglContext, fNum, normal, 9);
		
		// setup viewport and prespective
		GetWindowPortBounds(window, &rectPort);
		viewRect.size.width =(float)(rectPort.right - rectPort.left);
		viewRect.size.height =(float)(rectPort.bottom - rectPort.top);
		
		// setup viewport and prespective
		resizeGL(pContextInfo, viewRect); // forces projection matrix update
		
		//SetLighting(4);
		
		SetPort(portSave);
	}
	return err;
}

// ---------------------------------

// dump window data structures and OpenGL context and related data structures
OSStatus disposeGL(pRecContext pContextInfo)
{
	// release our data
	if ( pContextInfo != NULL )
	{
		if (pContextInfo->windowEDMUPP) { // dispose UPP for DM notifications
			DisposeDMExtendedNotificationUPP(pContextInfo->windowEDMUPP);
			pContextInfo->windowEDMUPP = NULL;
		}
		aglSetCurrentContext(NULL);
		aglSetDrawable(pContextInfo->aglContext, NULL);
		if (pContextInfo->aglContext) {
			aglDestroyContext(pContextInfo->aglContext);
			pContextInfo->aglContext = NULL;
		}
		if (pContextInfo->aglPixFmt) {
			aglDestroyPixelFormat(pContextInfo->aglPixFmt);
			pContextInfo->aglPixFmt = NULL;
		}
		if (pContextInfo->timer) {
			RemoveEventLoopTimer(pContextInfo->timer);
			pContextInfo->timer = NULL;
		}
		// dump font display list
		if (pContextInfo->boldFontList) {
			deleteFontGL(pContextInfo->boldFontList);
			pContextInfo->boldFontList = 0;
		}
		if (pContextInfo->regFontList) {
			deleteFontGL(pContextInfo->regFontList);
			pContextInfo->regFontList = 0;
		}
    if (pContextInfo->world) {
      //pContextInfo->ground = 0;
			delete pContextInfo->world;
      pContextInfo->world = 0;
    }
	}
	
	return noErr;
}

#pragma mark ---- OpenGL Drawing ----

#include "drawInfo.h" // source info for info output

// ---------------------------------

// draw text info
// note: this bitmap technique is not the speediest and one should use textures for font in more proformance critical code
static void drawInfo(pRecContext pContextInfo)
{	
	static float msgPresistance = 10.0f;
	//char cstr [256];
	GLint matrixMode, line = 1;
	GLboolean depthTest = glIsEnabled(GL_DEPTH_TEST);
	GLfloat height, width;
	
	if (!pContextInfo)
		return;
	
	height = pContextInfo->camera.viewHeight;
	width = pContextInfo->camera.viewWidth-200;
	
	glDisable(GL_DEPTH_TEST); // ensure text is not remove by deoth buffer test.
	//glEnable(GL_BLEND); // for text fading
	//glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA); // ditto
	
	// set orthograhic 1:1  pixel transform in local view coords
	glGetIntegerv(GL_MATRIX_MODE, &matrixMode);
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	glScalef(2.0 / width, -2.0 /  height, 1.0);
	glTranslatef(-width / 2.0, -height / 2.0, 0.0);

	glColor3f(0, 1.0, 0);
	glBegin(GL_LINE_LOOP);
	glVertex3f(0, 0, 0);
	glVertex3f(width, 0, 0);
	glVertex3f(width, height, 0);
	glVertex3f(0, height, 0);
	glEnd();
	
	if ((pContextInfo->controlShip) && (pContextInfo->gameState == kInGame) &&
						(pContextInfo->controlShip->isAlive() == true))
	{
		glColor4f(1.0, 1.0, 1.0, 0.5);
		glBegin(GL_LINES);
		glVertex3f(width/2-5, height/2-5, 0);
		glVertex3f(width/2-5, height/2+5, 0);
		glVertex3f(width/2+5, height/2+5, 0);
		glVertex3f(width/2+5, height/2-5, 0);
		glEnd();
	}
	
	// output strings
	glColor3f(1.0, 1.0, 1.0);

	// message string
	if (pContextInfo->message[0]) {
		float currDelta = getElapsedTime() - pContextInfo->msgTime;
		glColor4f(1.0, 1.0, 1.0,(msgPresistance - currDelta) * 0.1);
		glRasterPos3d(10, line++ * 12, 0); 
		drawCStringGL(pContextInfo->message, pContextInfo->boldFontList);
		if (currDelta > msgPresistance)
			pContextInfo->message[0] = 0;
	}
	// global error message
	if (gErrorMessage[0]) {
		float currDelta = getElapsedTime() - gErrorTime;
		glColor4f(1.0, 0.2, 0.2,(msgPresistance - currDelta) * 0.1);
		glRasterPos3d(10, line++ * 12, 0); 
		drawCStringGL(gErrorMessage, pContextInfo->boldFontList);
		if (currDelta > msgPresistance)
			gErrorMessage[0] = 0;
	}
	// reset orginal martices
	glPopMatrix(); // GL_MODELVIEW
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(matrixMode);
	if (depthTest)
		glEnable(GL_DEPTH_TEST);
	glReportError();
}

// ---------------------------------

// main OpenGL drawing function
void drawGL(pRecContext pContextInfo, Boolean swap)
{
	if (!pContextInfo || !(pContextInfo->aglContext))
		return;
	// ensure the context is current	
	if (!aglSetCurrentContext(pContextInfo->aglContext))
		aglReportError();
	
	glClear(GL_COLOR_BUFFER_BIT);

	// top right viewport
	glViewport(pContextInfo->camera.viewWidth-200, pContextInfo->camera.viewHeight-200, 200, 200);
	glClear(GL_DEPTH_BUFFER_BIT);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_BLEND);

	glMatrixMode(GL_PROJECTION); glLoadIdentity(); glMatrixMode(GL_MODELVIEW); glLoadIdentity();
	glColor3f(0, 1.0, 0); glBegin(GL_LINE_LOOP);
	glVertex3f(-1, 1, 0); glVertex3f(-1, -1, 0); glVertex3f(1, -1, 0); glVertex3f(1, 1, 0); glEnd();

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(-pContextInfo->zoom, pContextInfo->zoom, -pContextInfo->zoom, pContextInfo->zoom, -pContextInfo->zoom, pContextInfo->zoom);
	updateModelView(pContextInfo, true);
	switch (pContextInfo->controlShip?pContextInfo->controlShip->getOverviewpoint():kTop)
	{
		case kRear:
			pContextInfo->rearRadar->draw();
			break;
		case kSide:
			pContextInfo->sideRadar->draw();
			break;
		case kTop:
		default:
			pContextInfo->topRadar->draw();
			break;
	}
  if (pContextInfo->radarZoom)
    pContextInfo->radarZoom->draw();
	pContextInfo->world->drawOverview();
	
	// middle right viewport
	glViewport(pContextInfo->camera.viewWidth-200, pContextInfo->camera.viewHeight-400, 200, 200);
	glClear(GL_DEPTH_BUFFER_BIT);
	glDisable(GL_DEPTH_TEST);
	
	glMatrixMode(GL_PROJECTION); glLoadIdentity(); glMatrixMode(GL_MODELVIEW); glLoadIdentity();
	glColor3f(0, 1.0, 0); glBegin(GL_LINE_LOOP);
	glVertex3f(-1, 1, 0); glVertex3f(-1, -1, 0); glVertex3f(1, -1, 0); glVertex3f(1, 1, 0); glEnd();
	pContextInfo->status->draw();
	if (pContextInfo->controlShip)
		pContextInfo->controlShip->drawStatus();
	
	// bottom right viewport
	glViewport(pContextInfo->camera.viewWidth-200, pContextInfo->camera.viewHeight-600, 200, 200);
	glClear(GL_DEPTH_BUFFER_BIT);
	glDisable(GL_DEPTH_TEST);
	
	glMatrixMode(GL_PROJECTION); glLoadIdentity(); 	glOrtho(-1, 1, -1.0, 1, -1, 1);
	glMatrixMode(GL_MODELVIEW); glLoadIdentity();
	glColor3f(0, 1.0, 0); glBegin(GL_LINE_LOOP);
	glVertex3f(-1, 1, 0); glVertex3f(-1, -1.0, 0); glVertex3f(1, -1.0, 0); glVertex3f(1, 1, 0); glEnd();
	if (pContextInfo->level)
	pContextInfo->level->draw();
	if (pContextInfo->lives)
		pContextInfo->lives->draw();
	updateScoreTime(pContextInfo);
	if (pContextInfo->score)
		pContextInfo->score->draw();
	if (pContextInfo->ltime)
		pContextInfo->ltime->draw();

	
	// bottom-most right viewport
	glViewport(pContextInfo->camera.viewWidth-200, 0, 200, 100);
	glClear(GL_DEPTH_BUFFER_BIT);
	glDisable(GL_DEPTH_TEST);
	glMatrixMode(GL_PROJECTION); glLoadIdentity(); 	gluOrtho2D(-1, 1, 0, 1);
	glMatrixMode(GL_MODELVIEW); glLoadIdentity();
	pContextInfo->nssoftware->draw();

	glEnable(GL_DEPTH_TEST);
	glReportError();
	
	// main game viewport
	glViewport(0, 0, pContextInfo->camera.viewWidth-200, pContextInfo->camera.viewHeight);
	glClear(GL_DEPTH_BUFFER_BIT);
	//glEnable(GL_BLEND); // for text fading
	//glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA); // ditto
	updateProjection(pContextInfo);
	updateModelView(pContextInfo, false);

  glDisable(GL_LIGHTING);
	pContextInfo->world->draw();
	drawInfo(pContextInfo);
	
	if (swap)
		aglSwapBuffers(pContextInfo->aglContext);
	else
		glFlush();
}

#pragma mark ---- Carbon Timer ----

// per-window timer function, basic time based animation preformed here
static void timerContextCB(pRecContext pContextInfo)
{
	
	AbsoluteTime currTime = UpTime();
	double deltaTime =(float) AbsoluteDeltaToDuration(currTime, pContextInfo->time);
	
	pContextInfo->time = currTime;	// reset for next time interval
	if (0 > deltaTime)	// if negative microseconds
		deltaTime /= -1000000.0;
	else				// else milliseconds
		deltaTime /= 1000.0;
	if (deltaTime > 10.0) // skip pauses
	{
		printf("Elapsed time %1.6f (skipping)\n", deltaTime);
		return;
	}
	else {
		//printf("Elapsed time %1.6f\n", deltaTime);
		//showPressed();
		// 30 rotation per second, 1.5 velocity unit increase/second, .95 speed decrease per 1/30th second
		
		if (pContextInfo->gameState == kInGame)
		{
			double x = 1.0/128.0; // ~ 1/30*6 // do we really need to run this fast?
			while (x < deltaTime)
			{
				pContextInfo->world->stepTime(1.0/128.0); // physics run at a constant speed, (in game)
				x+=1.0/128.0;
			}
			pContextInfo->world->stepTime(1.0/128.0-x+deltaTime); // but adjust to current frame
		}
    else if (pContextInfo->gameState != kPaused)
      pContextInfo->world->stepTime(deltaTime); // run at a slower speed
    
		if ((pContextInfo->controlShip) && (pContextInfo->gameState == kInGame) &&
				(pContextInfo->controlShip->isAlive() == false))
		{
			point3d a(-.5, -.5, 1), b(.5, -1.0, 1);
			pContextInfo->world->addObject(new TextBox("CLICK TO CONTINUE", 17, a, b, 1000));
			pContextInfo->livesLeft--;
			if (pContextInfo->livesLeft >= 0)
      {
        //printf("Entering game state kBetweenLevels\n");
				pContextInfo->gameState = kBetweenLevels;
      }
			else {
        //printf("Entering game state kWaitingClick\n");
				pContextInfo->gameState = kWaitingClick;
				//pContextInfo->currentLevel = 0;
				CGDisplayShowCursor(CGMainDisplayID());
			}
		}
		if ((pContextInfo->controlShip) && (pContextInfo->gameState == kInGame) &&
				(pContextInfo->controlShip->isAlive() == true) && (pContextInfo->controlShip->passedLevel()))
		{
			pContextInfo->currentLevel++;
      SavePrefs(pContextInfo);
      //printf("ship says score is %d\n", pContextInfo->controlShip->getScore());
      //printf("General score goes from %d", pContextInfo->currentScore);
			pContextInfo->currentScore+=1000*pContextInfo->currentLevel+pContextInfo->controlShip->getScore();
      pContextInfo->controlShip->addToScore(-pContextInfo->controlShip->getScore());
      //printf(" to %d", pContextInfo->currentScore);
			if (pContextInfo->controlShip->getTime() < 30+pContextInfo->currentLevel*10)
      {
				pContextInfo->currentScore+=100*(30+pContextInfo->currentLevel*10-pContextInfo->controlShip->getTime());
        //printf(" to %d", pContextInfo->currentScore);
      }
      //printf("\n");
			if (pContextInfo->currentLevel < pContextInfo->ll->getNumLevels())
			{
				//point3d a(-.5, -.25, 1), b(.5, -1.0, 1);
				//pContextInfo->world->addObject(new TextBox("  PASSED LEVEL!  CLICK TO CONTINUE", 17, a, b, 1000));
				//pContextInfo->gameState = kBetweenLevels;
				loadLevelDescription(pContextInfo);
			}
			else {
				point3d a(-.5, .25, 1), b(.5, -.25, 1);
				pContextInfo->world->addObject(new TextBox("YOU PASSED ALL THE LEVELS!    CLICK TO CONTINUE     ", 26, a, b, 1000));
				//pContextInfo->currentLevel = 0;
        //printf("Entering game state kWaitingClick\n");
				pContextInfo->gameState = kWaitingClick;
				CGDisplayShowCursor(CGMainDisplayID());
			}
		}
    if (pContextInfo->gameState != kPaused)
      drawGL(pContextInfo, true); // required to do this directly to get animation during resize and drags
	}
}

// ---------------------------------

// timer callback bottleneck
static pascal void timerCB(EventLoopTimerRef inTimer, void* userData)
{
#pragma unused(inTimer)
	timerContextCB(GetCurrentContextInfo((WindowRef) userData)); // timer based update
}

// ---------------------------------

// get UPP for timer
static EventLoopTimerUPP getTimerUPP(void)
{
	static EventLoopTimerUPP	sTimerUPP = NULL;
	
	if (sTimerUPP == NULL)
		sTimerUPP = NewEventLoopTimerUPP(timerCB);
	
	return sTimerUPP;
}

#pragma mark ---- Carbon Event Handling ----

// intializes context conditions
static void initialConditions(pRecContext pContextInfo)
{
	BlockZero(pContextInfo, sizeof(recContext));
	resetCamera(&pContextInfo->camera);
	pContextInfo->info = kInfoState;
	pContextInfo->animate = kAnimateState;
	pContextInfo->drawCaps = 0;
	pContextInfo->drawHelp = 0;
	//pContextInfo->drawCursor = 1;
	pContextInfo->polygons = 0;
	pContextInfo->lines = 1;
	pContextInfo->points = 0;
	pContextInfo->showCredits = 0;
  //printf("Entering game state kAtMainMenu\n");
	pContextInfo->gameState = kAtMainMenu;
	pContextInfo->currentLevel = 0;
	pContextInfo->currentScore = 0;
	pContextInfo->livesLeft = 3;
	pContextInfo->pauseMessage = 0;
	pContextInfo->lighting = 4;
	pContextInfo->timer = NULL;
		
	pContextInfo->surface = 0; 
	pContextInfo->colorScheme = 4;
	pContextInfo->subdivisions = 64;
	pContextInfo->xyRatio = 1;
	
	pContextInfo->modeFSAA = 0;
  
  pContextInfo->world = new World();
	pContextInfo->controlShip = 0;
	point3d a(-.9, .9, 1), b(.9, -1.0, 1);
	pContextInfo->status = new TextBox("  shots  mines  speed  ", 23, a, b, 24*60*60);
	pContextInfo->sideRadar = new TextBox("RADAR (SIDE)", 20, a, b, 24*60*60);
	pContextInfo->topRadar = new TextBox("RADAR (TOP)", 20, a, b, 24*60*60);
	pContextInfo->rearRadar = new TextBox("RADAR (REAR)", 20, a, b, 24*60*60);
	a.y = 0+.75/2;
	b.y = 0-.75/2;
	pContextInfo->highScore = 0;
	b.y = -1.5;
	a.y = 1.0-.07;
	pContextInfo->level = new TextBox("LEVEL 1", 10, a, b, 24*60*60);
	a.y = .50-.07;
	pContextInfo->lives = new TextBox("3 LIVES", 10, a, b, 24*60*60);
	a.y = .00-.07;
	pContextInfo->score = new TextBox("0000000", 10, a, b, 24*60*60);
	a.y = -.5-.07;
	pContextInfo->ltime = new TextBox("0:00", 10, a, b, 24*60*60);
	a.y = .77;
	pContextInfo->nssoftware = new TextBox("WWW.NS-SOFTWARE.COM                      www.idevgames.com", 20, a, b, 24*60*60, false);
	pContextInfo->highScore = 0;
  pContextInfo->zoom = 128;
  pContextInfo->radarZoom = 0;
	pContextInfo->nameInitialsIndex = 0;
	pContextInfo->nameInitialsTextBox = 0;
	pContextInfo->hs = new HighScores();	
	pContextInfo->ll = new LevelLoader();
	setupMainMenu(pContextInfo);
}

// ---------------------------------

pRecContext GetCurrentContextInfo(WindowRef window)
{
	if (NULL == window)  // HID use this path
		window = FrontWindow();
	if (window)
		return(pRecContext) GetWRefCon(window);
	else
		return NULL;
}

// ---------------------------------

// move camera in z axis
static void mouseDolly(HIPoint location, pRecContext pContextInfo)
{
//	GLfloat dolly =(gDollyPanStartPoint[1] - location.y) * -pContextInfo->camera.viewPos.z / 300.0f;
//	pContextInfo->camera.viewPos.z += dolly;
//	if (pContextInfo->camera.viewPos.z == 0.0) // do not let z = 0.0
//		pContextInfo->camera.viewPos.z = 0.0001;
//	updateProjection(pContextInfo);  // update projection matrix
//	gDollyPanStartPoint[0] =(long)location.x;
//	gDollyPanStartPoint[1] =(long)location.y;
}

// ---------------------------------

// move camera in x/y plane
static void mousePan(HIPoint location, pRecContext pContextInfo)
{
//	GLfloat panX =(gDollyPanStartPoint[0] - location.x) /(900.0f / -pContextInfo->camera.viewPos.z);
//	GLfloat panY =(gDollyPanStartPoint[1] - location.y) /(900.0f / -pContextInfo->camera.viewPos.z);
//	pContextInfo->camera.viewPos.x -= panX;
//	pContextInfo->camera.viewPos.y -= panY;
//	gDollyPanStartPoint[0] =(long)location.x;
//	gDollyPanStartPoint[1] =(long)location.y;
}

// ---------------------------------

// Handles global non-system mouse events for GL windows as follows:
// primary button = object tumble(trackball)
// secondary button(or cntl-primary) = pan
// tertiary button(or option-primary) = dolly
// wheel = aperture

static OSStatus handleWindowMouseEvents(EventHandlerCallRef myHandler, EventRef event)
{
	WindowRef			window = NULL;
	pRecContext 		pContextInfo = NULL;
	OSStatus			result = eventNotHandledErr;
	UInt32 				kind;
	EventMouseButton	button = 0;
	HIPoint				location = {0.0f, 0.0f};
	UInt32				modifiers = 0;	
	long				wheelDelta = 0;		
	Rect 				rectPort;
	Rect r;
	
	kind = GetEventKind(event);
	// Mac OS X v10.1 and later
	// should this be front window???
	window = FrontWindow();
	//GetEventParameter(event, kEventParamWindowRef, typeWindowRef, NULL, sizeof(WindowRef), NULL, &window);
	if (window)
		pContextInfo =(pRecContext) GetWRefCon(window);
	if (!pContextInfo)
	{
		//printf("No appropriate context available\n");
		return result; // not an application GLWindow so do not process(there is an exception)
	}
	GetWindowPortBounds(window, &rectPort);
	GetWindowBounds(window, kWindowContentRgn, &r);
	
	result = CallNextEventHandler(myHandler, event);	
	if (eventNotHandledErr == result) 
	{ // only handle events not already handled(prevents wierd resize interaction)
		switch(kind) {
			case kEventMouseWheelMoved: 
				GetEventParameter(event, kEventParamMouseWheelDelta, typeLongInteger, NULL, sizeof(long), NULL, &wheelDelta);
				if (wheelDelta)
				{
					if (wheelDelta < 0)
          {
            if (pContextInfo->zoom < 512)
            {
              pContextInfo->zoom *= 2;
              updateRadarZoom(pContextInfo);
            }
          }
          else if (wheelDelta > 0) {
            if (pContextInfo->zoom > 2)
            {
              pContextInfo->zoom /= 2;
              updateRadarZoom(pContextInfo);
            }
          }            
				}
					break;
			case kEventMouseDown:				
				switch (pContextInfo->gameState)
				{
					case kAtMainMenu:
						GetEventParameter(event, kEventParamMouseLocation, typeHIPoint, NULL, sizeof(HIPoint), NULL, &location);	// Mac OS X v10.1 and later
						handleMainMenuClick(pContextInfo, location, r);
						return noErr;
						break;
					case kWaitingClick:
						handleWaitingClick(pContextInfo);
						return noErr;
						break;
					case kEnterHigh:
						GetEventParameter(event, kEventParamMouseLocation, typeHIPoint, NULL, sizeof(HIPoint), NULL, &location);	// Mac OS X v10.1 and later
						handleEnterHighClick(pContextInfo, location, r);
						return noErr;
						break;
					case kBetweenLevels:
						handleBetweenLevelClick(pContextInfo, r);
						break;
					case kPaused:
						handlePausedClick();
						return noErr;
						break;
					case kInGame:
						GetEventParameter(event, kEventParamMouseButton, typeMouseButton, NULL, sizeof(EventMouseButton), NULL, &button);
						GetEventParameter(event, kEventParamKeyModifiers, typeUInt32, NULL, sizeof(UInt32), NULL, &modifiers);
						handleGameClick(pContextInfo, button, modifiers);
						break;
				}
			case kEventMouseDragged:
			case kEventMouseUp:
			case kEventMouseMoved:
				if ((pContextInfo->gameState == kInGame) || (pContextInfo->gameState == kBetweenLevels))
				{
          GetEventParameter(event, kEventParamMouseDelta, typeHIPoint, NULL, sizeof(HIPoint), NULL, &location);	// Mac OS X v10.1 and later
          handleGameMouseMovement(pContextInfo, location, r);
        }
				break;
		}
		result = noErr;
	}	
	return noErr;
}

// ---------------------------------


// key input handler
static OSStatus handleKeyInput(EventHandlerCallRef myHandler, EventRef event, Boolean keyDown, void* userData)
{
	enum { kH = 0x04,   // toggle draw help
		kC = 0x08,   // toggle draw caps
		kLfSqBracket = 0x21,   // color scheme down
		kRtSqBracket = 0x1E,   // color scheme up
		kSemi = 0x29,   // decrease subdivisions
		kQuote = 0x27,   // increase subdivisions
		kF = 0x03,   // toggle fill
		kP = 0x23,   // toggle points
		kW = 0x0D,   // toggle wire
		kL = 0x25,   // next lighting model(wraps)
		kBackSlash = 0x2A,   // next shape(wraps)
		kQuestion = 0x11,   // toggle credits
		kM = 0x2E,	// cycle multi-sample
		kS = 0x01,	// smooth
		kA = 0x00,
		kR = 0x0F,
		kD = 0x02,
		kLeftArrow = 0x7b,
		kRightArrow = 0x7c,
		kUpArrow = 0x7e,
		kDownArrow = 0x7d,
		kSpace = 0x31,
		kEscape = 0x35
	};

	WindowRef window =(WindowRef) userData;
	OSStatus result = eventNotHandledErr;
	
	result = CallNextEventHandler(myHandler, event);	
	if (eventNotHandledErr == result) { 
		pRecContext pContextInfo = GetCurrentContextInfo(window);
		if (pContextInfo) {
			UInt32 keyCode;
			Rect rectPort = {0,0,0,0};
			GetEventParameter(event, kEventParamKeyCode, typeUInt32, NULL, sizeof(UInt32), NULL, &keyCode);
			// handle keyboard input here
			if (keyDown) {
				//printf("hit 0x%lx\n", keyCode);
				switch(keyCode) {
          case kUpArrow:
            if (pContextInfo->zoom < 512)
            {
              pContextInfo->zoom *= 2;
              updateRadarZoom(pContextInfo);
            }
            break;
          case kDownArrow:
            if (pContextInfo->zoom > 2)
            {
              pContextInfo->zoom /= 2;
              updateRadarZoom(pContextInfo);
            }
            break;
					case kEscape:
						if (pContextInfo->gameState == kInGame)
						{
              //printf("Entering game state kPaused\n");
							pContextInfo->gameState = kPaused;
							point3d a(-.5, 1-1.0/24.0, 1), b(.5, -1.0/24.0, 1);
							pContextInfo->world->addObject(pContextInfo->pauseMessage = new TextBox("        GAME PAUSED     PRESS ESCAPE TO CONTINUE", 24, a, b, 24*60*60));
							pContextInfo->world->stepTime(0);
              drawGL(pContextInfo, true);
							CGDisplayShowCursor(CGMainDisplayID());
						}
						else if (pContextInfo->gameState == kPaused)
						{
              //printf("Entering game state kInGame\n");
							pContextInfo->gameState = kInGame;
							if (pContextInfo->pauseMessage)
              {
								pContextInfo->pauseMessage->destroy();
                pContextInfo->pauseMessage = 0;
              }
							first = true;
							CGDisplayHideCursor(CGMainDisplayID());
							Rect r;
							CGPoint point;
							GetWindowBounds(window, kWindowContentRgn, &r);
							point.x = r.left+(r.right-r.left)/2;
							point.y = r.top+(r.bottom-r.top)/2;
							/*CGDisplayErr=*/
							//printf("2Resetting mouse to (%1.2f, %1.2f)\n", point.x, point.y);
							CGDisplayMoveCursorToPoint(CGMainDisplayID(), point);
						}
						//else
							//CGDisplayShowCursor(CGMainDisplayID());
						//pContextInfo->drawCursor = 1 - pContextInfo->drawCursor;
						break;
					case kH:
						// toggle help
						pContextInfo->drawHelp = 1 - pContextInfo->drawHelp;
						GetWindowPortBounds(window, &rectPort);
						InvalWindowRect(window, &rectPort);
						result = noErr;
						break;
					case kC:
						// toggle caps
						pContextInfo->drawCaps = 1 - pContextInfo->drawCaps;
						GetWindowPortBounds(window, &rectPort);
						InvalWindowRect(window, &rectPort);
						result = noErr;
						break;
					case kQuestion: // credits
						break;
					case kF: // fill
						break;
					case kP: // points
						break;
					case kR: // radar
						if (pContextInfo->controlShip)
							pContextInfo->controlShip->rotateOverviewpoint();
						break;
					case kL: // lighting
						break;
					case kLfSqBracket: // next lower color scheme
						break;
					case kRtSqBracket: // next higher color scheme
						break;
					case kSemi: // next lower subdivision setting
						break;
					case kQuote:
						break;
					case kBackSlash: // next higher surface
						break;
#ifdef kUseMultiSample
					case kM: // multisample rotate
						pContextInfo->modeFSAA =(pContextInfo->modeFSAA + 1) %(kFSAANice + 1);
						switch(pContextInfo->modeFSAA) {
							case kFSAAOff:
								glDisable(GL_MULTISAMPLE_ARB);
								break;
							case kFSAAFast:
								glEnable(GL_MULTISAMPLE_ARB);
								glHint(GL_MULTISAMPLE_FILTER_HINT_NV, GL_FASTEST);
								break;
							case kFSAANice:
								glEnable(GL_MULTISAMPLE_ARB);
								glHint(GL_MULTISAMPLE_FILTER_HINT_NV, GL_NICEST);
								break;
						}
            GetWindowPortBounds(window, &rectPort);
						InvalWindowRect(window, &rectPort);
#endif
						break;
				}
			}
		}
	}
	return noErr;//result;
}

// ---------------------------------

// window update event handling
static void handleWindowUpdate(WindowRef window)
{
	// set port for QD drawing
	GrafPtr portSave;
	GetPort(&portSave);
	SetPort(GetWindowPort(window));
	drawGL(GetCurrentContextInfo(window), true); // in this case just draw content
	SetPort(portSave);
}

// ---------------------------------

void createNewWindow(void)
{
	EventHandlerRef ref;
	EventTypeSpec list[] = { { kEventClassWindow, kEventWindowCollapsing },
	{ kEventClassWindow, kEventWindowCollapsed },
	{ kEventClassWindow, kEventWindowShown },
	{ kEventClassWindow, kEventWindowActivated },
	{ kEventClassWindow, kEventWindowClose },
	{ kEventClassWindow, kEventWindowDrawContent },
	{ kEventClassWindow, kEventWindowBoundsChanged },
	{ kEventClassWindow, kEventWindowZoomed },
	{ kEventClassKeyboard, kEventRawKeyDown },
//	{ kEventClassKeyboard, kEventRawKeyRepeat },
//	{ kEventClassKeyboard, kEventRawKeyUp } 
  };
	
	
	WindowRef window = NULL;
	pRecContext pContextInfo =(pRecContext) NewPtrClear(sizeof(recContext)); // memory for window record
	initialConditions(pContextInfo);
	CreateWindowFromNib(nibRef, CFSTR("MainWindow"), &window); // build window
	if (window)
	{
		Rect r;
		SetPortWindowPort(window);
		SetWRefCon(window,(long) pContextInfo); // point to the window record in the ref con of the window
    GetWindowPortBounds(window, &r);
    InvalWindowRect(window, &r);
		InstallWindowEventHandler(window, gWinEvtHandler, GetEventTypeCount(list), list,(void*)window, &ref); // add event handler
		SelectWindow(window);
		ShowWindow(window);
    pContextInfo->time = UpTime();
    InstallEventLoopTimer(GetCurrentEventLoop(), 0, kEventDurationSecond/30, getTimerUPP(),(void *) window, &pContextInfo->timer);
    LoadPrefs(pContextInfo);
	}
}

// ---------------------------------

// window event handler
static pascal OSStatus windowEvtHndlr(EventHandlerCallRef myHandler, EventRef event, void* userData)
{
#pragma unused(userData)
	UInt32				eClass;
	UInt32 				kind;
	WindowRef			window = NULL;
	pRecContext			pContextInfo = NULL;
	Rect				rectPort = {0,0,0,0};
	CGRect 				viewRect = {{0.0f, 0.0f}, {0.0f, 0.0f}};
	OSStatus			result = eventNotHandledErr;
	eClass = GetEventClass(event);
	kind = GetEventKind(event);
	
	
	switch(eClass) {
		case kEventClassKeyboard:
			switch(kind) {
				case kEventRawKeyDown:
					result = handleKeyInput(myHandler, event, true, userData);
					break;
				case kEventRawKeyUp:
					result = handleKeyInput(myHandler, event, false, userData);
					break;
				case kEventRawKeyRepeat:
					result = handleKeyInput(myHandler, event, true, userData);
					break;
			}
			break;
		case kEventClassWindow:
			GetEventParameter(event, kEventParamDirectObject, typeWindowRef, NULL, sizeof(WindowRef), NULL, &window);
			switch(kind) {
				case kEventWindowCollapsing:
					pContextInfo =(pRecContext) GetWRefCon(window);

          if (pContextInfo->gameState == kInGame)
          {
            //printf("Entering game state kPaused\n");
            pContextInfo->gameState = kPaused;
            point3d a(-.5, 1-1.0/24.0, 1), b(.5, -1.0/24.0, 1);
            pContextInfo->world->addObject(pContextInfo->pauseMessage = new TextBox("        GAME PAUSED     PRESS ESCAPE TO CONTINUE", 24, a, b, 60*60*24));
            pContextInfo->world->stepTime(0);
            drawGL(pContextInfo, true);
            CGDisplayShowCursor(CGMainDisplayID());
          }
          else {
              drawGL(GetCurrentContextInfo(window), false); // in this case just draw content
          }
					GetWindowPortBounds(window, &rectPort);
					CompositeGLBufferIntoWindow( pContextInfo->aglContext, &rectPort, window);
					result = UpdateCollapsedWindowDockTile(window);
					break;
				case kEventWindowActivated: // called on click activation and initially
					handleWindowUpdate(window);
					break;
				case kEventWindowDrawContent:
					handleWindowUpdate(window);
					break;
				case kEventWindowClose: // called when window is being closed(close box)
					HideWindow(window);
					pContextInfo =(pRecContext) GetWRefCon(window); // get window info
					disposeGL(pContextInfo);
					DisposePtr((Ptr) pContextInfo);
					SetWRefCon(window, NULL);
					QuitApplicationEventLoop();
					break;
				case kEventWindowShown: // called on initial show(not on un-minimize)
					buildGL(window);
					if (window == FrontWindow())
						SetUserFocusWindow(window);
					GetWindowPortBounds(window, &rectPort);
					InvalWindowRect(window, &rectPort);
					break;
				case kEventWindowBoundsChanged: // called for resize and moves(drag)
					GetWindowPortBounds(window, &rectPort); // get port rect
					pContextInfo =(pRecContext) GetWRefCon(window); // get window info	
					GetWindowPortBounds(window, &rectPort);
					viewRect.size.width =(float)(rectPort.right - rectPort.left);
					viewRect.size.height =(float)(rectPort.bottom - rectPort.top);
					// if we are not animating and shrinking, force update
					if (pContextInfo && !pContextInfo->timer && 
							((pContextInfo->camera.viewHeight > viewRect.size.height) ||
							(pContextInfo->camera.viewWidth > viewRect.size.width))) {
						resizeGL(pContextInfo, viewRect);
						handleWindowUpdate(window); // must force immediate update to get live resize
					} else
						resizeGL(pContextInfo, viewRect);					
					break;
				case kEventWindowZoomed: // called when user clicks on pContextInfo->zoom button
					GetWindowPortBounds(window, &rectPort);
					viewRect.size.width =(float)(rectPort.right - rectPort.left);
					viewRect.size.height =(float)(rectPort.bottom - rectPort.top);
					resizeGL(pContextInfo, viewRect);
					break;
			}
				break;
	}
	return result;
}

// ---------------------------------

// application level event handler
static pascal OSStatus appEvtHndlr(EventHandlerCallRef myHandler, EventRef event, void* userData)
{
#pragma unused(myHandler, userData)
	OSStatus result = eventNotHandledErr;
	Rect rectPort;
	pRecContext pContextInfo = NULL;
	WindowRef window = FrontWindow();
	UInt32 eClass = GetEventClass(event);
	UInt32 kind = GetEventKind(event);
	HICommand command;
	
	// gss: use current context call
	if (window) { // do we have a window?
		GetWindowPortBounds(window, &rectPort); // get bounds for later inval
		pContextInfo =(pRecContext) GetWRefCon(window); // get window info
	}
	
	switch(eClass) {
		case kEventClassMouse:
			handleWindowMouseEvents(myHandler, event);
			break;
		case kEventClassCommand:
			switch(kind) {
				case kEventProcessCommand:
					GetEventParameter(event, kEventParamDirectObject, kEventParamHICommand, NULL, sizeof(command), NULL, &command); // get command
					switch(command.commandID) {
						//case 'abou':
							//CreateWindowFromNib(nibRef, CFSTR("AboutWindow"), &window); // build window
              //if (window)
              //{
              //	SelectWindow(window);
              //	ShowWindow(window);
              //}
							//	break;
            case 'musi':
            {
              gMusic = 1-gMusic;
              MenuRef mr = GetMenuHandle(501);
              CheckMenuItem(mr, 1, gMusic);
              SavePrefs(pContextInfo);
            }
              break;
            case 'snd ':
            {
              gSound = 1-gSound;
              MenuRef mr = GetMenuHandle(501);
              CheckMenuItem(mr, 2, gSound);
              SavePrefs(pContextInfo);
            }
              break;
            case 'tail':
            {
              gDetail = 1-gDetail;
              MenuRef mr = GetMenuHandle(501);
              CheckMenuItem(mr, 3, gDetail);
              SavePrefs(pContextInfo);
            }
            break;
            case 'invt':
            {
              gInvert = 1-gInvert;
              MenuRef mr = GetMenuHandle(501);
              CheckMenuItem(mr, 4, gInvert);
              SavePrefs(pContextInfo);
            }
              break;
						case 'KEYS':
							CreateWindowFromNib(nibRef, CFSTR("Keys"), &window); // build window
							if (window)
							{
								SelectWindow(window);
								ShowWindow(window);
							}
								break;
						case 'endg':
							if (pContextInfo->controlShip)
							{
                //printf("Entering game state kInGame\n");
                pContextInfo->gameState = kInGame;
								pContextInfo->controlShip->destroy();
								//pContextInfo->controlShip = 0;
								pContextInfo->livesLeft = 0;
							}
							break;
						case 'neww':
              pContextInfo->currentLevel = 0;
            case 'CONT':
							first = true;
							resetNewGame(pContextInfo);
							CGDisplayHideCursor(CGMainDisplayID());
							CGPoint point;
							Rect r;
							GetWindowBounds(window, kWindowContentRgn, &r);
							point.x = r.right+(r.left-r.right)/2;
							point.y = r.top+(r.bottom-r.top)/2;
							/*CGDisplayErr=*/
							//printf("3Resetting mouse to (%1.2f, %1.2f)\n", point.x, point.y);
							CGDisplayMoveCursorToPoint(CGMainDisplayID(), point);
							loadLevelDescription(pContextInfo);
							//SetPortWindowPort(window);
							InvalWindowRect(window, &rectPort);
							//createNewWindow();
							result = noErr;
							break;
						case 'clsw':
							if (window) {
								HideWindow(window);
								disposeGL(GetCurrentContextInfo(window));  // dump our structures
								DisposeWindow(window); // if it exists dump it
							}
							break;
						case 'quit':
							//memory will get dumped on exit, other callbacks/timers are for this process only
							break;
						case 'gogo':
							if (pContextInfo) {  // have window info
								pContextInfo->animate = 1 - pContextInfo->animate;
								if (pContextInfo->animate && !pContextInfo->timer) {
									pContextInfo->time = UpTime();
									InstallEventLoopTimer(GetCurrentEventLoop(), 0, kEventDurationSecond/30, getTimerUPP(),(void *) window, &pContextInfo->timer);
								} else if (!pContextInfo->animate && pContextInfo->timer) {
									RemoveEventLoopTimer(pContextInfo->timer);
									pContextInfo->timer = NULL;
								}
							}
							break;
						case 'info':
							if (pContextInfo) {  // have window info
								pContextInfo->info = 1 - pContextInfo->info;
								GetWindowPortBounds(window, &rectPort);
								InvalWindowRect(window, &rectPort);
							}
							break;
					}
						break;
				case kEventCommandUpdateStatus:
					GetEventParameter(event, kEventParamDirectObject, kEventParamHICommand, NULL, sizeof(command), NULL, &command); // get command
					switch(command.commandID) {
						case 'clsw':
							if (pContextInfo)
								EnableMenuItem(GetMenuHandle(kMainMenu), kCloseMenuItem);
							else 
								DisableMenuItem(GetMenuHandle(kMainMenu), kCloseMenuItem);
							break;
						case 'gogo':
							if (pContextInfo) {
								EnableMenuItem(GetMenuHandle(kMainMenu), kAnimateMenuItem);
								CheckMenuItem(GetMenuHandle(kMainMenu), kAnimateMenuItem, pContextInfo->animate);
							} else { 
								DisableMenuItem(GetMenuHandle(kMainMenu), kAnimateMenuItem);
								CheckMenuItem(GetMenuHandle(kMainMenu), kAnimateMenuItem, kAnimateState);
							}
							break;
						case 'info':
							if (pContextInfo) {
								EnableMenuItem(GetMenuHandle(kMainMenu), kInfoMenuItem);
								CheckMenuItem(GetMenuHandle(kMainMenu), kInfoMenuItem, pContextInfo->info);
							} else { 
								DisableMenuItem(GetMenuHandle(kMainMenu), kInfoMenuItem);
								CheckMenuItem(GetMenuHandle(kMainMenu), kInfoMenuItem, kInfoState);
							}
							break;
					}
						result = noErr;
					break;
			}
			break;
	}
	return result;
}

bool isPressed(unsigned short k)
// k =  any keyboard scan code, 0-127
{
	unsigned long km[4];
	
	GetKeys( (long *) km );
	return (km[k>>5]>>(k&0x1F))&0x1;
}

void showPressed()
{
	unsigned long km[4];
	GetKeys( (long *) km );
	for (int x = 0; x < 4; x++)
		for (int y = 0; (y < 32)&&(km[x]); y++)
		{
			if (km[x]&0x1)
				printf("%d\t", x*32+y);
			km[x]>>=1;
		}
}

void resetNewGame(pRecContext pContextInfo)
{
	pContextInfo->currentScore = 0;
	pContextInfo->livesLeft = 3;
	//pContextInfo->currentLevel = 6; // just for now...
}

void loadLevelDescription(pRecContext pContextInfo)
{
  //printf("Entering game state kBetweenLevels\n");
	pContextInfo->gameState = kBetweenLevels;
	pContextInfo->ll->loadLevelDescription(pContextInfo->currentLevel, pContextInfo->world);
  pContextInfo->controlShip = 0;
}

void loadLevel(pRecContext pContextInfo)
{
	int which = pContextInfo->currentLevel;
  pContextInfo->controlShip = pContextInfo->ll->loadLevel(which, pContextInfo->world);
  //printf("Entering game state kInGame\n");
  pContextInfo->gameState = kInGame;
  pContextInfo->zoom = 128;
  updateRadarZoom(pContextInfo);
  MenuRef mr = GetMenuHandle(500);
  if (mr)
  {
    EnableMenuCommand(mr, 'endg');
    DisableMenuCommand(mr, 'neww');
    DisableMenuCommand(mr, 'CONT');
  }
  
	char level[16];
	sprintf(level, "LEVEL %d", which);
	//pContextInfo->level->destroy();
	delete pContextInfo->level;
	pContextInfo->level = 0;
	point3d a(-.9, .84, 1), b(.9, -1.0, 1);
	a.y = 1-.07;
	pContextInfo->level = new TextBox(level, 10, a, b, 24*60*60);		

	//pContextInfo->lives->destroy();
	delete pContextInfo->lives;
	pContextInfo->lives = 0;
	a.y =.5-.07;
	sprintf(level, "%d LIVES", pContextInfo->livesLeft);
	pContextInfo->lives = new TextBox(level, 10, a, b, 24*60*60);
}

int getPrefValue(CFStringRef key)
{
  Boolean found;
	int result = -1;
  result = CFPreferencesGetAppIntegerValue(key, CFSTR("com.NSSoftware.SVRacer"), &found);
	if (found)
	{
    //printf("Found results %d in prefs\n", result);
    return result;
	}
  //printf("Didn't Found results in prefs\n");
	return -1;
}

void setPrefValue(CFStringRef key, int value)
{
	CFPropertyListRef val = CFNumberCreate(NULL, kCFNumberIntType, &value);
	CFPreferencesSetAppValue(key, val, CFSTR("com.NSSoftware.SVRacer"));
	CFRelease(val);
}

void SavePrefs(pRecContext pContextInfo)
{
	CFStringRef music = CFSTR("Music");
	CFStringRef sound = CFSTR("Sound");
	CFStringRef quality = CFSTR("HighDetail");
	CFStringRef lastLevel = CFSTR("LastLevel");
  CFStringRef invert = CFSTR("Invert");
	
	setPrefValue(music, gMusic);
	setPrefValue(sound, gSound);
	setPrefValue(quality, gDetail);
	setPrefValue(lastLevel, pContextInfo->currentLevel);
  setPrefValue(invert, gInvert);
  
	CFPreferencesAppSynchronize(kCFPreferencesCurrentApplication);
}

void LoadPrefs(pRecContext pContextInfo)
{
  int tmp;
	CFStringRef music = CFSTR("Music");
	CFStringRef sound = CFSTR("Sound");
	CFStringRef quality = CFSTR("HighDetail");
	CFStringRef lastLevel = CFSTR("LastLevel");
  CFStringRef invert = CFSTR("Invert");
  
	if ((tmp = getPrefValue(music)) != -1)
		gMusic = tmp;
	if ((tmp = getPrefValue(sound)) != -1)
		gSound = tmp;
	if ((tmp = getPrefValue(quality)) != -1)
		gDetail = tmp;
	if ((tmp = getPrefValue(lastLevel)) != -1)
		pContextInfo->currentLevel = tmp;
	if ((tmp = getPrefValue(invert)) != -1)
    gInvert = tmp;

  MenuRef mr = GetMenuHandle(501);
  CheckMenuItem(mr, 1, gMusic);
  CheckMenuItem(mr, 2, gSound);
  CheckMenuItem(mr, 3, gDetail);  
  CheckMenuItem(mr, 4, gInvert);  
  if (pContextInfo->currentLevel != 0)
  {
    MenuRef mr = GetMenuHandle(500);
    if (mr)
    {
      EnableMenuCommand(mr, 'CONT');
    }
  }
}  

void updateRadarZoom(pRecContext pContextInfo)
{
  char text[21];
  if (pContextInfo->radarZoom)
    pContextInfo->radarZoom->destroy();
	point3d a(-.9, -.72, 1), b(.9, -1.0, 1);
  if (pContextInfo->zoom >= 128)
    sprintf(text, "               %1.2fx", 128.0/(float)pContextInfo->zoom);
  else
    sprintf(text, "                %3dx", 128/pContextInfo->zoom);
  pContextInfo->radarZoom = new TextBox(text, 20, a, b, 24*60*60);
}

void updateScoreTime(pRecContext pContextInfo)
{
	static int score = 0, time = 0;
	char text[16];
	point3d a(-.9, .84, 1), b(.9, -1.0, 1);
  if (pContextInfo->gameState != kInGame)
    return;
	if ((pContextInfo->controlShip) && (pContextInfo->controlShip->getTime() != time))
	{
		time = pContextInfo->controlShip->getTime();
		a.y = -.5-.07;
		sprintf(text, "%d:%d%d", time/60, (time%60)/10, time%10);
		//pContextInfo->ltime->destroy();
		delete pContextInfo->ltime;
		pContextInfo->ltime = 0;
		pContextInfo->ltime = new TextBox(text, 10, a, b, 24*60*60);
	}
	if ((pContextInfo->controlShip) && (pContextInfo->controlShip->getScore()+pContextInfo->currentScore != score))
	{
		score = pContextInfo->controlShip->getScore()+pContextInfo->currentScore;
		a.y =0-.07;
		sprintf(text, "%d%d%d%d%d%d%d", (score/1000000)%10, (score/100000)%10,
						(score/10000)%10, (score/1000)%10, (score/100)%10, (score/10)%10,
						(score)%10);
		//pContextInfo->score->destroy();
		delete pContextInfo->score;
		pContextInfo->score = 0;
		pContextInfo->score = new TextBox(text, 10, a, b, 24*60*60);
	}
}

void setupMainMenu(pRecContext pContextInfo)
{
  MenuRef mr = GetMenuHandle (500);
  if (mr)
  {
    DisableMenuCommand(mr, 'endg');
    EnableMenuCommand(mr, 'neww');
    if (pContextInfo->currentLevel != 0)
      EnableMenuCommand(mr, 'CONT');
    else
      DisableMenuCommand(mr, 'CONT');
  }

  Stars *s;
	double numItems = 5;
	double dir = (random()%2)?(-1):(1);
	pContextInfo->world->addObject(s = new Stars(8000));
	s->startRotating(dir);
	pContextInfo->world->addObject(s = new Stars(2000));
	s->startRotating(dir*.9);
	for (int x = 0; x < 5; x++)
	{
		point3d a(-.5, 1-((x+1.0)*2.0/numItems)+1/numItems+1.0/13.0, 1), b(.5, -1, 1);
		switch (x)
		{
			case 0:
				pContextInfo->world->addObject(new TextBox("   NEW GAME", 13, a, b, 24*60*60)); // we dare you to leave the game up for a day!
				break;
			case 1:
				pContextInfo->world->addObject(new TextBox("  ABOUT GAME", 13, a, b, 24*60*60));
				break;
			case 2:
				pContextInfo->world->addObject(new TextBox("   CONTROLS", 13, a, b, 24*60*60));
				break;
			case 3:
				pContextInfo->world->addObject(new TextBox(" HIGH  SCORES", 13, a, b, 24*60*60));
				break;
			case 4:
				pContextInfo->world->addObject(new TextBox("     QUIT", 13, a, b, 24*60*60));
				break;
		}
	}
}

void handleMainMenuClick(pRecContext pContextInfo, HIPoint location, Rect r)
{
//	printf("Mouse at (%1.2f, %1.2f), window (%d, %d, %d, %d)\n",
//				 location.x, location.y, r.left, r.top, r.right, r.bottom);
	
	if ((location.x > r.right-200) && (location.y > r.bottom-50))
	{ // clicked on NS-SOFTARE!
		ICInstance inst;
		Str255 url = "\phttp://www.idevgames.com/";
		long start = 0, end = url[0];
		ICStart(&inst, 'SPRC');
		ICLaunchURL(inst, "\p", &url[1], url[0], &start, &end);
		ICStop(inst);						
	}
	else if ((location.x > r.right-200) && (location.y > r.bottom-100))
	{ // clicked on NS-SOFTARE!
		ICInstance inst;
		Str255 url = "\phttp://www.ns-software.com";
		long start = 0, end = url[0];
		ICStart(&inst, 'SPRC');
		ICLaunchURL(inst, "\p", &url[1], url[0], &start, &end);
		ICStop(inst);						
	}
	if (location.x < r.left+700)
	{
		if (location.y < r.top+(r.bottom-r.top)/5)
		{
			first = true;
      pContextInfo->currentLevel = 0;
			resetNewGame(pContextInfo);
			CGDisplayHideCursor(CGMainDisplayID());
			CGPoint point;
			point.x = r.left+(r.right-r.left)/2;
			point.y = r.top+(r.bottom-r.top)/2;
			//printf("4Resetting mouse to (%1.2f, %1.2f)\n", point.x, point.y);
			CGDisplayMoveCursorToPoint(CGMainDisplayID(), point);
			loadLevelDescription(pContextInfo);
			//SetPortWindowPort(window);
			//InvalWindowRect(window, &rectPort);
		}
		else if (location.y < r.top+2*(r.bottom-r.top)/5)
		{
      // ABOUT
      pContextInfo->ll->loadLevelDescription(-2, pContextInfo->world);
      //printf("Entering game state kWaitingClick\n");
			pContextInfo->gameState = kWaitingClick;
		}
		else if (location.y < r.top+3*(r.bottom-r.top)/5)
		{
      // KEYS
      pContextInfo->ll->loadLevelDescription(-1, pContextInfo->world);
      //printf("Entering game state kWaitingClick\n");
			pContextInfo->gameState = kWaitingClick;
		}
		else if (location.y < r.top+4*(r.bottom-r.top)/5)
		{
			// saveHighScore
      Stars *s;
      pContextInfo->controlShip = 0;
      pContextInfo->pauseMessage = 0;
      pContextInfo->pauseMessage = 0;
      pContextInfo->world->reset();
			pContextInfo->world->addObject(s = new Stars(4000));
      s->startRotating();
			pContextInfo->highScore = 0;
			pContextInfo->world->addObject(pContextInfo->hs->getHighScoreList());
			pContextInfo->currentScore = 0;
      //printf("Entering game state WaitingClick\n");
			pContextInfo->gameState = kWaitingClick;
		}
		else {
			QuitApplicationEventLoop();
			//quit!
		}
	}	
}

void handleWaitingClick(pRecContext pContextInfo)
{
	pContextInfo->world->reset();
  pContextInfo->pauseMessage = 0;
	pContextInfo->controlShip = 0;
	if (pContextInfo->hs->isHighScore(pContextInfo->currentScore))
	{
		pContextInfo->nameInitialsText[0] = 0;
		pContextInfo->nameInitialsIndex = 0;
	
    //printf("Entering game state kEnterHigh\n");
		pContextInfo->gameState = kEnterHigh;
		point3d a(-.9, 1.05/2, 1), b(.9, -1.05/2, 1);						
		pContextInfo->world->addObject(pContextInfo->highScore=new TextBox(" YOU GOT A HIGH SCORE                            A  B  C  D  E  F  G  H  I  J  K  L  M  N  O  P  Q  R  S  T  U  V  W  X  Y  Z  !  _  .                         RUB   OK ", 24, a, b, 24*60*60));
		a.y = -1.05/2-.15;
		a.x = -.1125;
		pContextInfo->world->addObject(new TextBox("___", 24, a, b, 24*60*60));
    pContextInfo->nameInitialsIndex = 0;
	}
	else {
    //printf("Entering game state kAtMainMenu\n");
		pContextInfo->gameState = kAtMainMenu;
		setupMainMenu(pContextInfo);
	}	
}

void handleEnterHighClick(pRecContext pContextInfo, HIPoint location, Rect r)
{
	char what[32] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ!_.";
	bool redo = false;
	double x = location.x-r.left;
	double y = location.y-r.top;
	x = x/350-1;
	y = 1-y/350;
	//printf("Approximated (%1.2f, %1.2f)\n", x, y);
	if ((x > -.9) && (x < .9) && (y < .225) && (y > -.525))
	{
		//printf("Letter row %d, column %d\n", (int)(5-(y+.525)/.15), (int)((x+.9)/.225));
		int loc = 8*(int)((double)((5-(y+.525)/.15)))+(int)((double)((x+.9)/.225));
		if ((loc >= 0) && (loc < 29))
		{
			if (pContextInfo->nameInitialsIndex < 3)
			{
				pContextInfo->nameInitialsText[pContextInfo->nameInitialsIndex] = what[loc];
				pContextInfo->nameInitialsIndex++;
				pContextInfo->nameInitialsText[pContextInfo->nameInitialsIndex] = 0;
				redo = true;
			}
			//printf("Hit %c %d\n", what[loc%32], loc);
		}
		if (loc == 37)
		{
			//printf("RUB\n");
			if (pContextInfo->nameInitialsIndex > 0)
			{
				pContextInfo->nameInitialsIndex--;
				pContextInfo->nameInitialsText[pContextInfo->nameInitialsIndex] = 0;
				redo = true;
			}
		}
		if (loc == 39)
		{
			//printf("OK\n");
			pContextInfo->hs->addHighScore(pContextInfo->nameInitialsText, pContextInfo->currentScore);
			// saveHighScore
      pContextInfo->controlShip = 0;
      pContextInfo->pauseMessage = 0;
			pContextInfo->world->reset();
			pContextInfo->world->addObject(new Stars(4000));
			pContextInfo->highScore = 0;
			//setupMainMenu(pContextInfo);
			pContextInfo->world->addObject(pContextInfo->hs->getHighScoreList());
			pContextInfo->nameInitialsTextBox = 0;
			pContextInfo->currentScore = 0;
      //printf("Entering game state kWaitingClick\n");
			pContextInfo->gameState = kWaitingClick;
		}
	}
	if (redo)
	{
		if (pContextInfo->nameInitialsTextBox)
			pContextInfo->nameInitialsTextBox->destroy();
		point3d a(-.1125, -1.05/2-.13, 1), b(.9, -1, 1);
		pContextInfo->nameInitialsTextBox = new TextBox(pContextInfo->nameInitialsText, 24, a, b, 24*60*60);
		pContextInfo->world->addObject(pContextInfo->nameInitialsTextBox);
	}
}

void handleBetweenLevelClick(pRecContext pContextInfo, Rect r)
{
  if (pContextInfo->currentLevel == 0)
  {
    pContextInfo->currentLevel++;
    loadLevelDescription(pContextInfo);
    return;
  }
	first = true;
	//CGDisplayHideCursor(CGMainDisplayID());
	CGPoint point;
	point.x = r.left+(r.right-r.left)/2;
	point.y = r.top+(r.bottom-r.top)/2;
	/*CGDisplayErr=*/
	//printf("5Resetting mouse to (%1.2f, %1.2f)\n", point.x, point.y);
	CGDisplayMoveCursorToPoint(CGMainDisplayID(), point);
	loadLevel(pContextInfo);
}

void handlePausedClick()
{
}

void handleGameClick(pRecContext pContextInfo, EventMouseButton button, UInt32 modifiers)
{
	if ((button == kEventMouseButtonSecondary) || (modifiers & controlKey))
	{
		if (pContextInfo->controlShip)
			pContextInfo->controlShip->dropMine();
	}
	else {
		if (pContextInfo->controlShip)
			pContextInfo->controlShip->shoot();
	}
}

void handleGameMouseMovement(pRecContext pContextInfo, HIPoint location, Rect r)
{
	//printf("Mouse offset (%1.2f, %1.2f)\n", location.x, location.y);
	CGPoint point;
	point.x = r.left+(r.right-r.left)/2;
	point.y = r.top+(r.bottom-r.top)/2;
	//printf("1Resetting mouse to (%1.2f, %1.2f)\n", point.x, point.y);
	CGDisplayMoveCursorToPoint(CGMainDisplayID(), point);

  if (pContextInfo->gameState == kBetweenLevels)
    return;

	if ((first) && ((location.x != 0) || (location.y != 0)))
	{
		first = false;
		//printf("Breaking first time\n");
		return;
	}
	
  //showPressed();
  //printf("\n");
  if (isPressed(32))
  {
    //printf("Using directional modifier\n");
    if (fabs(location.x) > fabs(location.y))
      location.y = 0;
    else
      location.x = 0;
  }
	float rotation[4] = {0.0f, 0.0f, 0.0f, 0.0f};
  if (gInvert)
    rotation[0] = -location.y/6;
  else
    rotation[0] = location.y/6;
	rotation[1] = 1;
	if ((pContextInfo->controlShip) && (rotation[0] != 0))
		pContextInfo->controlShip->addRotation(rotation);
	//addToRotationTrackball(rotation, pContextInfo->fRot);
	rotation[0] = location.x/6;
	rotation[1] = 0; rotation[2] = 1;
	if ((pContextInfo->controlShip)  && (rotation[0] != 0))
		pContextInfo->controlShip->addRotation(rotation);
}

#pragma mark ==== main ====

int main(int argc, char* argv[])
{
	OSStatus		err;
	EventHandlerRef	ref;
	EventTypeSpec	list[] = { { kEventClassCommand,  kEventProcessCommand },
	{ kEventClassCommand,  kEventCommandUpdateStatus },
	{ kEventClassMouse, kEventMouseMoved},
	{ kEventClassMouse, kEventMouseDown },// handle trackball functionality globaly because there is only a single user
	{ kEventClassMouse, kEventMouseUp }, 
	{ kEventClassMouse, kEventMouseDragged },
	{ kEventClassMouse, kEventMouseWheelMoved }
	};

	ProcessSerialNumber psn = { 0, kCurrentProcess };
	
	gStartTime = UpTime(); // get app start time
	
	// Create a Nib reference passing the name of the nib file(without the .nib extension)
	// CreateNibReference only searches into the application bundle.
	err = CreateNibReference(CFSTR("main"), &nibRef);
	require_noerr( err, CantGetNibRef );
	err = SetMenuBarFromNib(nibRef, CFSTR("MainMenu"));
	require_noerr( err, CantSetMenuBar );
	gEvtHandler = NewEventHandlerUPP(appEvtHndlr);
	InstallApplicationEventHandler(gEvtHandler, GetEventTypeCount(list), list, 0, &ref);
	gWinEvtHandler = NewEventHandlerUPP(windowEvtHndlr);
	
	getCurrentCaps();
	
	// ensure we know when display configs are changed
	gConfigEDMUPP = NewDMExtendedNotificationUPP(handleConfigDMEvent); // for display change notification
	DMRegisterExtendedNotifyProc(gConfigEDMUPP, NULL, NULL, &psn);
	
  RegisterHelpBook();
	//StartHIDInput();
	
	createNewWindow();
	// Call the event loop
	RunApplicationEventLoop();
	
	// Exiting...
	
	if (gConfigEDMUPP) { // dispose UPP for DM notifications
		DisposeDMExtendedNotificationUPP(gConfigEDMUPP);
		gConfigEDMUPP = NULL;
	}
	
	if (gDisplayCaps) { // dispose diaplsy capabilities info
		free(gDisplayCaps);
		gDisplayCaps = NULL;
	}
	
	//CantCreateWindow:
CantSetMenuBar:
CantGetNibRef:
		
		// We don't need the nib reference anymore.
		DisposeNibReference(nibRef);
	
	return err;
}


void RegisterHelpBook()
{
	CFBundleRef myApplicationBundle;
  CFURLRef myBundleURL;
  FSRef myBundleRef;
  OSStatus err = noErr;
  
  myApplicationBundle = NULL;
  myBundleURL = NULL;
  
  myApplicationBundle = CFBundleGetMainBundle();         //1
  if (myApplicationBundle == NULL) {err = fnfErr; }
	else {
		myBundleURL = CFBundleCopyBundleURL(myApplicationBundle);         //2
		if (myBundleURL == NULL) {err = fnfErr;}
		else {
			if (!CFURLGetFSRef(myBundleURL, &myBundleRef)) err = fnfErr;              //3
      
			if (err == noErr) err = AHRegisterHelpBook(&myBundleRef);                //4
		}
	}
	
}
