/*
 (MIT License)
 
 Copyright (c) 2004 NS Software
 
 Permission is hereby granted, free of charge, to any person obtaining a copy of this
 software and associated documentation files (the "Software"), to deal in the Software
 without restriction, including without limitation the rights to use, copy, modify,
 merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
 permit persons to whom the Software is furnished to do so, subject to the following
 conditions:
 
 The above copyright notice and this permission notice shall be included in all copies
 or substantial portions of the Software.
 
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
 INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
 OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
 */
/*
 *  vector3d.cpp
 *  Carbon OpenGL
 *
 *  Created by Nathan Sturtevant on Wed Mar 24 2004.
 *  Copyright (c) 2004 NS Software.
 *
 */

#include <math.h>
#include "vector3d.h"

vector3d::vector3d(GLfloat a, GLfloat b, GLfloat c)
:x(a), y(b), z(c)
{ }

float vector3d::squaredLength()
{
	return x*x + y*y + z*z;
}

float vector3d::length()
{
	return sqrt(x*x + y*y + z*z);
}

void vector3d::normalize()
{
  double length;
  
  length = sqrt(x*x + y*y + z*z);
  if (length != 0) {
    x /= length;
    y /= length;
    z /= length;
  } else {
    x = 0;
    y = 0;
    z = 0;
  }
}

vector3d vector3d::crossProduct(vector3d v)
{
  vector3d n(y * v.z - z * v.y,
             z * v.x - x * v.z,
             x * v.y - y * v.x);
  return n;
}